==> randovania-2.6.1.tar.gz: randovania-2.6.1/randovania/generator/elevator_distributor.py <==
import copy
from collections import defaultdict
from random import Random
from typing import List, Dict, Optional, Tuple

from randovania.game_description.area_location import AreaLocation
from randovania.game_description.node import TeleporterNode
from randovania.game_description.teleporter import Teleporter
from randovania.game_description.world_list import WorldList

CUSTOM_NAMES_FOR_ELEVATORS = {
    # Great Temple
    408633584: "Temple Transport Emerald",
    2399252740: "Temple Transport Violet",
    2556480432: "Temple Transport Amber",

    # Temple Grounds to Great Temple
    1345979968: "Sanctuary Quadrant",
    1287880522: "Agon Quadrant",
    2918020398: "Torvus Quadrant",

    # Temple Grounds to Areas
    1660916974: "Agon Gate",
    2889020216: "Torvus Gate",
    3455543403: "Sanctuary Gate",

    # Agon
    1473133138: "Agon Entrance",
    2806956034: "Agon Portal Access",
    3331021649: "Agon Temple Access",

    # Torvus
    1868895730: "Torvus Entrance",
    3479543630: "Torvus Temple Access",
    3205424168: "Lower Torvus Access",

    # Sanctuary
    3528156989: "Sanctuary Entrance",
    900285955: "Sanctuary Spider side",
    3145160350: "Sanctuary Vault side",
}


class ElevatorHelper:
    teleporter: Teleporter
    destination: AreaLocation
    connected_elevator: Optional["ElevatorHelper"]

    def __init__(self, teleporter: Teleporter, destination: AreaLocation):
        self.teleporter = teleporter
        self.destination = destination
        self.connected_elevator = None

    @property
    def world_asset_id(self):
        return self.teleporter.world_asset_id

    @property
    def area_asset_id(self):
        return self.teleporter.area_asset_id

    @property
    def instance_id(self):
        return self.teleporter.instance_id

    def connect_to(self, other: "ElevatorHelper"):
        self.destination = other.teleporter.area_location
        other.destination = self.teleporter.area_location
        self.connected_elevator = other
        other.connected_elevator = self

    @property
    def area_location(self):
        return AreaLocation(self.world_asset_id, self.area_asset_id)


def try_randomize_elevators(rng: Random,
                            echoes_elevators: Tuple[ElevatorHelper, ...],
                            ) -> List[ElevatorHelper]:
    elevator_database: List[ElevatorHelper] = list(echoes_elevators)
    assert len(elevator_database) % 2 == 0

    elevator_list = copy.copy(elevator_database)
    elevators_by_world: Dict[int, List[ElevatorHelper]] = defaultdict(list)
    for elevator in elevator_list:
        elevators_by_world[elevator.world_asset_id].append(elevator)

    while elevator_list:
        source_elevators: List[ElevatorHelper] = max(elevators_by_world.values(), key=len)
        target_elevators: List[ElevatorHelper] = [
            elevator
            for elevator in elevator_list
            if elevator not in source_elevators
        ]
        source_elevator = source_elevators[0]
        target_elevator = rng.choice(target_elevators)

        source_elevator.connect_to(target_elevator)

        elevators_by_world[source_elevator.world_asset_id].remove(source_elevator)
        elevators_by_world[target_elevator.world_asset_id].remove(target_elevator)
        elevator_list.remove(source_elevator)
        elevator_list.remove(target_elevator)

    # TODO
    list3 = copy.copy(elevator_database)
    celevator_list3 = [list3[0]]
    while list3:
        celevator_list1 = []
        for celevator1 in celevator_list3:
            index = 0
            while index < len(list3):
                celevator2 = list3[index]
                if (celevator2.world_asset_id == celevator1.world_asset_id
                        or celevator2.area_asset_id == celevator1.destination.area_asset_id):
                    celevator_list1.append(celevator2)
                    list3.remove(celevator2)
                else:
                    index += 1
        if celevator_list1:
            celevator_list3 = celevator_list1
        else:
            # Randomization failed
            return try_randomize_elevators(rng, echoes_elevators)

    return elevator_database


def two_way_elevator_connections(rng: Random,
                                 elevator_database: Tuple[ElevatorHelper, ...],
                                 between_areas: bool
                                 ) -> Dict[int, AreaLocation]:
    if len(elevator_database) % 2 != 0:
        raise ValueError("Two-way elevator shuffle, but odd number of elevators to shuffle.")
    if between_areas:
        elevator_database = try_randomize_elevators(rng, elevator_database)
    else:
        elevators = list(elevator_database)
        rng.shuffle(elevators)
        while elevators:
            elevators.pop().connect_to(elevators.pop())

    return {
        elevator.instance_id: elevator.connected_elevator.area_location
        for elevator in elevator_database
    }


def one_way_elevator_connections(rng: Random,
                                 elevator_database: Tuple[ElevatorHelper, ...],
                                 target_locations: List[AreaLocation],
                                 replacement: bool,
                                 ) -> Dict[int, AreaLocation]:
    target_locations.sort()
    rng.shuffle(target_locations)

    def _create_target():
        if replacement:
            return rng.choice(target_locations)
        else:
            return target_locations.pop()

    return {
        elevator.instance_id: _create_target()
        for elevator in elevator_database
    }


def create_elevator_database(world_list: WorldList,
                             all_teleporters: List[Teleporter],
                             ) -> Tuple[ElevatorHelper, ...]:
    """
    Creates a tuple of Elevator objects, exclude those that belongs to one of the areas provided.
    :param world_list:
    :param all_teleporters: Set of teleporters to use
    :return:
    """
    all_helpers = [
        ElevatorHelper(Teleporter(world.world_asset_id, area.area_asset_id, node.teleporter_instance_id),
                       node.default_connection)

        for world, area, node in world_list.all_worlds_areas_nodes
        if isinstance(node, TeleporterNode)
    ]
    return tuple(
        helper
        for helper in all_helpers
        if helper.teleporter in all_teleporters
    )

==> randovania-3.1.2.tar.gz: randovania-3.1.2/randovania/generator/elevator_distributor.py <==
import copy
from collections import defaultdict
from random import Random
from typing import List, Dict, Optional, Tuple

from randovania.game_description.world.area_location import AreaLocation
from randovania.game_description.game_patches import ElevatorConnection
from randovania.game_description.world.node import TeleporterNode
from randovania.game_description.world.teleporter import Teleporter
from randovania.game_description.world.world_list import WorldList


class ElevatorHelper:
    teleporter: Teleporter
    destination: AreaLocation
    connected_elevator: Optional["ElevatorHelper"]

    def __init__(self, teleporter: Teleporter, destination: AreaLocation):
        self.teleporter = teleporter
        self.destination = destination
        self.connected_elevator = None

    @property
    def world_asset_id(self):
        return self.teleporter.world_asset_id

    @property
    def area_asset_id(self):
        return self.teleporter.area_asset_id

    @property
    def instance_id(self):
        return self.teleporter.instance_id

    def connect_to(self, other: "ElevatorHelper"):
        self.destination = other.teleporter.area_location
        other.destination = self.teleporter.area_location
        self.connected_elevator = other
        other.connected_elevator = self

    @property
    def area_location(self):
        return AreaLocation(self.world_asset_id, self.area_asset_id)


def try_randomize_elevators(rng: Random,
                            echoes_elevators: Tuple[ElevatorHelper, ...],
                            ) -> List[ElevatorHelper]:
    elevator_database: List[ElevatorHelper] = list(echoes_elevators)
    assert len(elevator_database) % 2 == 0

    elevator_list = copy.copy(elevator_database)
    elevators_by_world: Dict[int, List[ElevatorHelper]] = defaultdict(list)
    for elevator in elevator_list:
        elevators_by_world[elevator.world_asset_id].append(elevator)

    while elevator_list:
        source_elevators: List[ElevatorHelper] = max(elevators_by_world.values(), key=len)
        target_elevators: List[ElevatorHelper] = [
            elevator
            for elevator in elevator_list
            if elevator not in source_elevators
        ]
        source_elevator = source_elevators[0]
        target_elevator = rng.choice(target_elevators)

        source_elevator.connect_to(target_elevator)

        elevators_by_world[source_elevator.world_asset_id].remove(source_elevator)
        elevators_by_world[target_elevator.world_asset_id].remove(target_elevator)
        elevator_list.remove(source_elevator)
        elevator_list.remove(target_elevator)

    # TODO
    list3 = copy.copy(elevator_database)
    celevator_list3 = [list3[0]]
    while list3:
        celevator_list1 = []
        for celevator1 in celevator_list3:
            index = 0
            while index < len(list3):
                celevator2 = list3[index]
                if (celevator2.world_asset_id == celevator1.world_asset_id
                        or celevator2.area_asset_id == celevator1.destination.area_asset_id):
                    celevator_list1.append(celevator2)
                    list3.remove(celevator2)
                else:
                    index += 1
        if celevator_list1:
            celevator_list3 = celevator_list1
        else:
            # Randomization failed
            return try_randomize_elevators(rng, echoes_elevators)

    return elevator_database


def two_way_elevator_connections(rng: Random,
                                 elevator_database: Tuple[ElevatorHelper, ...],
                                 between_areas: bool
                                 ) -> ElevatorConnection:
    if len(elevator_database) % 2 != 0:
        raise ValueError("Two-way elevator shuffle, but odd number of elevators to shuffle.")
    if between_areas:
        elevator_database = try_randomize_elevators(rng, elevator_database)
    else:
        elevators = list(elevator_database)
        rng.shuffle(elevators)
        while elevators:
            elevators.pop().connect_to(elevators.pop())

    return {
        elevator.teleporter: elevator.connected_elevator.area_location
        for elevator in elevator_database
    }


def one_way_elevator_connections(rng: Random,
                                 elevator_database: Tuple[ElevatorHelper, ...],
                                 target_locations: List[AreaLocation],
                                 replacement: bool,
                                 ) -> ElevatorConnection:
    target_locations.sort()
    rng.shuffle(target_locations)

    def _create_target():
        if replacement:
            return rng.choice(target_locations)
        else:
            return target_locations.pop()

    return {
        elevator.teleporter: _create_target()
        for elevator in elevator_database
    }


def create_elevator_database(world_list: WorldList,
                             all_teleporters: List[Teleporter],
                             ) -> Tuple[ElevatorHelper, ...]:
    """
    Creates a tuple of Elevator objects, exclude those that belongs to one of the areas provided.
    :param world_list:
    :param all_teleporters: Set of teleporters to use
    :return:
    """
    all_helpers = [
        ElevatorHelper(node.teleporter, node.default_connection)

        for world, area, node in world_list.all_worlds_areas_nodes
        if isinstance(node, TeleporterNode)
    ]
    return tuple(
        helper
        for helper in all_helpers
        if helper.teleporter in all_teleporters
    )
